
/*
 * process_model.cc
 * Copyright (C) 1999 by John Heidemann
 * $Id: process_model.cc,v 1.16 1999/05/21 04:10:19 johnh Exp $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdlib.h>  // atoi
#include <iostream.h>

#include <unistd.h>
#include <time.h>

#include <stl.h>
#include <functional>

#include "process_model.hh"
#include "main.hh"


bool
process_model_ordering::operator()(process_model *a, process_model *b) const
{
	// For some reason we get called with a null...
	// if we do reorders() in age().
	// Recover from that.
	if (!a || !b)
		return false;
	if (a->exponential_age_ != b->exponential_age_)
		return a->exponential_age_ < b->exponential_age_;
	if (a->size_.last() != b->size_.last())
		return a->size_.last() < b->size_.last();
	return a->pid_ < b->pid_;
}


process_model::process_model_order_t process_model::orders_;

process_model::process_model(int pid) :
	order_(orders_.end()),
	pid_(pid),
	found_(true)
{
	age(time(NULL), true);
}

process_model::~process_model()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (order_ != orders_.end())
		orders_.erase(order_);
	order_ = orders_.end();
	delete pv_;
}

void
process_model::reorder()
{
	// optimize for not moving
	process_model_order_t::iterator prev_order_(order_);
	if (order_ != orders_.end()) {
		--prev_order_;
		orders_.erase(order_);
	};
	order_ = orders_.insert(prev_order_, this);
}

process_model *
process_model::superior()
{
	if (order_ == orders_.end())
	    return NULL;
	// I don't know how to check for moving our iterator
	// off the front.  I assumed it would return end(),
	// but maybe not.  The next check is paranoia.
	if (order_ == orders_.begin())
	    return NULL;
	process_model_order_t::iterator superior(order_);
	--superior;
	return superior == orders_.end() ? NULL : *superior;
}

void
process_model::age(time_t now, bool reinit)
{
	if (reinit || utime_.changed() || stime_.changed()) {
		last_run_time_ = now;
		age_increment_ = 4;
		exponential_age_ = 1;
		reorder();
	} else if (now > next_age_time_) {
		age_increment_ *= 2;
		exponential_age_++;
		reorder();
	} else return;

	next_age_time_ = last_run_time_ + age_increment_;
}

int
process_model::cmd_hashify(const char *cmd)
{
	// Hash function is the same as Tcl's.
	// see: generic/tclHash.c::HashString for details.
	// (Basically, hash*9 is good because it's fast and keeps
	// old values in both high bits and low bits.)
	int hash = 0;
	for (const unsigned char *cp = (const unsigned char*)cmd; *cp; cp++)
		hash += (hash << 3) + *cp;
	return hash;
}

void
process_model::set_cmd(const char *cmd)
{
	cmd_hash_.tick(cmd_hashify(cmd));
	if (cmd_hash_.changed())
		cmd_ = cmd;
}

void
process_model::dump()
{
	cout << int(this) << " "<< pid_ << " " << uid_ << " " << cmd_.c_str() << " " <<
		endl;
}
